﻿using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.DotProlog.Code;
using OpcLabs.DotProlog.Errors;
using OpcLabs.DotProlog.Execution;
using OpcLabs.DotProlog.InputOutput;
using OpcLabs.DotProlog.Terms;
using OpcLabs.DotProlog.Test;

namespace DotPrologTest.Execution
{
    [TestClass]
    // ReSharper disable once InconsistentNaming
    public class Call_1_Test : ExecutionTest
    {
        #region Static Private Methods

        // ReSharper disable once SuggestBaseTypeForParameter
        static private void Consult([NotNull] IProgram program)
        {
            IDatabase database = program.Database;

            database.Assert(ParseClause("b(X) :- Y = (write(X), X), call(Y)."));

            database.Assert(ParseClause("a(1)."));
            database.Assert(ParseClause("a(2)."));
        }

        #endregion



        #region Test Methods
        // ReSharper disable InconsistentNaming

        [TestMethod]
        public void Test_7_8_3_3_Errors_a()
        {
            using (IEngine engine = new Engine())
            {
                Term goal = Parse("call(G). ");
                TestError(engine, goal, ErrorTerms.InstantiationError);
            }
        }

        [TestMethod]
        public void Test_7_8_3_3_Errors_b()
        {
            using (IEngine engine = new Engine())
            {
                Term goal = Parse("call(`str`). ");
                TestError(engine, goal, ErrorTerms.TypeError(ValidTypes.Callable, new PString("str")));
            }
        }

        [TestMethod]
        public void Test_7_8_3_3_Errors_c()
        {
            using (IEngine engine = new Engine())
            {
                Term goal = Parse("call(1). ");
                TestError(engine, goal, ErrorTerms.TypeError(ValidTypes.Callable, new PInteger(1)));
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_1()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                Test(engine, Parse("call(!). "), Success(), "Expected success.");
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_2()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                Test(engine, Parse("call(fail). "), Failure(), "Expected failure.");
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_3()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                Test(engine, Parse("call((fail,X)). "), Failure(), "Expected failure.");
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_4()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                Test(engine, Parse("call( (fail, call(1)) ). "), Failure(), "Expected failure.");
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_5()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);
                TestError(engine, Parse("b(_). "), ErrorTerms.InstantiationError);
                var output = stringBuilder.ToString();
                Trace.WriteLine(output);
                StringAssert.StartsWith(output, "_");
                Assert.IsTrue(long.TryParse(output.Mid(1), out long result));
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_6()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);
                // Cor.2
                TestError(engine, Parse("b(3). "), ErrorTerms.TypeError(ValidTypes.Callable, Parse("write(3),3. ")));
                var output = stringBuilder.ToString();
                Assert.AreEqual("", output);
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_7()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);

                Term goal = Parse("Z = !, call( (Z=!, a(X), Z) ). ", out IDictionary<string, Variable> variableNameDictionary);
                Term expectedX = new PInteger(1);
                Term expectedZ = Atom.Add("!");
                var expected = new VariableDictionary
                {
                    { variableNameDictionary["X"], expectedX },
                    { variableNameDictionary["Z"], expectedZ }
                };
                Test(engine, goal, Success(expected));
                //FrontEnd.PrintResults(engine.Call(goal), variableNameDictionary);
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_8()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);

                Term goal = Parse("call( (Z=!, a(X), Z) ). ", out IDictionary<string, Variable> variableNameDictionary);
                var expected1 = new VariableDictionary
                {
                    { variableNameDictionary["X"], new PInteger(1) },
                    { variableNameDictionary["Z"], Atom.Add("!") }
                };
                var expected2 = new VariableDictionary
                {
                    { variableNameDictionary["X"], new PInteger(2) },
                    { variableNameDictionary["Z"], Atom.Add("!") }
                };
                Test(engine, goal, Success(expected1, expected2));
                //FrontEnd.PrintResults(engine.Call(goal), variableNameDictionary);
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_9()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);
                TestError(engine, Parse("call((write(3), X)). "), ErrorTerms.InstantiationError);
                var output = stringBuilder.ToString();
                Assert.AreEqual("3", output);
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_10()
        {
            var stringBuilder = new StringBuilder();
            using (IEngine engine = new Engine { Environment = { CurrentOutputStream = TextStream.Create(stringBuilder) } })
            {
                Consult(engine.Program);
                TestError(engine, Parse("call((write(3), call(1))). "), ErrorTerms.TypeError(ValidTypes.Callable, new PInteger(1)));
                var output = stringBuilder.ToString();
                Assert.AreEqual("3", output);
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_11()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                TestError(engine, Parse("call(X). "), ErrorTerms.InstantiationError);
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_12()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                TestError(engine, Parse("call(1). "), ErrorTerms.TypeError(ValidTypes.Callable, new PInteger(1)));
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_13()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                TestError(engine, Parse("call((fail, 1)). "), ErrorTerms.TypeError(ValidTypes.Callable, new NTerm(",") {"fail", 1}));
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_14()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                TestError(engine, Parse("call((write(3), 1)). "), ErrorTerms.TypeError(ValidTypes.Callable,
                    Parse("write(3), 1.")));
            }
        }

        [TestMethod]
        public void Test_7_8_3_4_Examples_15()
        {
            using (IEngine engine = new Engine())
            {
                Consult(engine.Program);
                TestError(engine, Parse("call((1;true)). "), ErrorTerms.TypeError(ValidTypes.Callable,
                    Parse("1;true.")));
            }
        }

        // ReSharper restore InconsistentNaming
        #endregion
    }
}
