﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Runtime.InteropServices;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    // ReSharper disable once InconsistentNaming
    public class IEasyDAClientTest_Com_BrowsePath : IEasyDAClientTestBase
    {
        public IEasyDAClientTest_Com_BrowsePath([NotNull] Func<IEasyDAClient> createClientFunc)
            : base(createClientFunc)
        {
        }

        protected void ReadItemAsserts(DAVtq actualVtq, object expectedValue)
        {
            Assert.IsNotNull(actualVtq);

            Assert.IsTrue(actualVtq.Quality.IsGood);
            Assert.AreEqual(DAStatusChoice.GoodNonspecific, actualVtq.Quality.StatusBitField);
            Assert.AreEqual(DALimitChoice.LimitOk, actualVtq.Quality.LimitBitField);

            Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
            Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
            Assert.IsTrue(actualVtq.Timestamp <= DateTime.Now);

            Assert.AreEqual(expectedValue, actualVtq.Value);
        }

        protected void ReadItemTest([NotNull] ServerDescriptor serverDescriptor, [NotNull] DAItemDescriptor itemDescriptor,
                                    [CanBeNull] object expectedValue)
        {
            var target = CreateClient();
            DAVtq actualVtq;
            try
            {
                actualVtq = target.ReadItem(serverDescriptor, itemDescriptor);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            ReadItemAsserts(actualVtq, expectedValue);
        }

        protected void ReadItemTest([NotNull] string machineName, [NotNull] string serverClass, [NotNull] string itemId,
            [NotNull] object expectedValue)
        {
            var target = CreateClient();
            DAVtq actualVtq = target.ReadItem(machineName, serverClass, itemId);
            ReadItemAsserts(actualVtq, expectedValue);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_BrowsePath()
        {
            ReadItemTest(
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(null, OpcLabs_KitServer.BrowsePaths.Constant_I4),
                0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_BrowsePath_Unknown()
        {
            var serverDescriptor = new ServerDescriptor("", OpcLabs_KitServer.ProgId);
            var itemDescriptor = new DAItemDescriptor(null, OpcLabs_KitServer.BrowsePaths.UnknownBrowsePath);
            var target = CreateClient();
            try
            {
                target.ReadItem(serverDescriptor, itemDescriptor);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }

                Exception innerException = opcException.InnerException;
                Assert.IsNotNull(innerException);

                Exception baseException = innerException.GetBaseException();
                Assert.IsNotNull(baseException);
                AssertUtilities.WriteLine($"baseException: {baseException}");
                Assert.IsInstanceOfType(baseException, typeof(ExtendedCOMException));
                var extendedComException = (ExtendedCOMException)baseException;
                AssertUtilities.WriteLine($"extendedComException.QualifiedName: {extendedComException.QualifiedName}");
                AssertUtilities.WriteLine($"extendedComException.GetErrorId(): {extendedComException.GetErrorId()}");
                // ReSharper disable once CommentTypo
                Assert.AreEqual(unchecked((int)0xC004903D), extendedComException.ErrorCode);   // OCK_E_ITEM_DATA_ID_NOT_FOUND
                Assert.AreEqual("http://opcfoundation.org/DataAccess/", extendedComException.QualifiedName.Namespace);
                Assert.AreEqual("E_FAIL.0xC004903D", extendedComException.QualifiedName.Name);
                return;
            }
            Assert.Fail();
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueTest_BrowsePath()
        {
            var target = CreateClient();
            object actualValue;
            try
            {
                actualValue = target.ReadItemValue(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new DAItemDescriptor(null, OpcLabs_KitServer.BrowsePaths.Constant_I4));
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            Assert.AreEqual(0, actualValue);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueTest_BrowsePath()
        {
            var target = CreateClient();
            try
            {
                target.WriteItemValue<bool>(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new DAItemDescriptor(null, OpcLabs_KitServer.BrowsePaths.Register_BOOL),
                    false);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest_BrowsePath()
        {
            var target = CreateClient();

            DAPropertyElementCollection actualResultCollection;
            try
            {
                actualResultCollection = target.BrowseProperties(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new DANodeDescriptor(null, OpcLabs_KitServer.BrowsePaths.Constant_I4));
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.IsNotNull(actualResultCollection);
            Assert.IsTrue(actualResultCollection.Count >= 12);

            //TODO:

        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetPropertyValueTest_BrowsePath()
        {
            var target = CreateClient();

            object actualValue;
            try
            {
                actualValue = target.GetPropertyValue(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new DANodeDescriptor(null, OpcLabs_KitServer.BrowsePaths.Constant_I4),
                    DAPropertyIds.DataType);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.AreEqual((short)VarTypes.I4, actualValue);

            //TODO:

        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest_BrowsePath()
        {
            IEasyDAClientTest_Com.SubscribeItemTest(
                CreateClient(), 
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(null, OpcLabs_KitServer.BrowsePaths.Incrementing_1s),
                unsubscribeAllItems: false);
        }
    }
}
