﻿
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Threading;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    partial class UAInteroperabilityWorkshopTest
    {
        #region Test Methods

        [TestMethod]
        public void Group006_SubscriptionServices_Header____________________()
        {
        }

        /// <summary>
        /// [obsolete] Create empty Subscription
        /// </summary>
        [TestMethod]
        public void Group006_SubscriptionServices_Test01_EmptySubscription()
        {
            ClientNotSupported();
        }

        /// <summary>
        /// Modify an existing subscription by varying subscription properties (e.g. publishingInterval and keepAlive etc.) and verify subscription behavior.
        /// </summary>
        // IMPROVE: this test does not yet recognize that the server may revise the sampling rate
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group006_SubscriptionServices_Test02_ModifySubscription()
        {
            TestSection.Arrange();
            int samplingInterval1 = DynamicSamplingInterval1;
            int samplingInterval2 = DynamicSamplingInterval2;
            const float tolerance = 0.6f;
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;
            AssertUtilities.WriteLine($"Sampling interval 1: {samplingInterval1}");
            AssertUtilities.WriteLine($"Sampling interval 2: {samplingInterval2}");
            AssertUtilities.WriteLine($"Tolerance: {tolerance}");

            UANodeDescriptor nodeDescriptor = DynamicNode1;
            AssertUtilities.WriteLine($"Node descriptor: {nodeDescriptor}");

            TestSection.Act();
            using (var client = new EasyUAClient
            {
                Isolated = true,
                IsolatedParameters = { SessionParameters = { EnableStatusSubscription = !SupportsJustOneSubscription}}
            })
            {
                var stopwatch = new Stopwatch();
                long[] lastElapsedMilliseconds = { -1, -1 };
                long[] intervalSum = { 0, 0 };
                int[] intervalCount = { 0, 0 };
                var synchronize = new object();
                client.DataChangeNotification += (sender, args) =>
                {
                    long elapsedMilliseconds = stopwatch.ElapsedMilliseconds;
                    int index = -1;
                    // ReSharper disable once ConvertIfStatementToSwitchStatement
                    // ReSharper disable once PossibleNullReferenceException
                    if (args.Arguments.MonitoringParameters.SamplingInterval == samplingInterval1)
                        index = 0;
                    else if (args.Arguments.MonitoringParameters.SamplingInterval == samplingInterval2)
                        index = 1;
                    if (index == -1)
                        return;

                    AssertUtilities.WriteLine($"@{elapsedMilliseconds} data change {index + 1}: {args}");
                    lock (synchronize)
                    {
                        if (lastElapsedMilliseconds[index] != -1)
                        {
                            intervalSum[index] += elapsedMilliseconds - lastElapsedMilliseconds[index];
                            intervalCount[index]++;
                        }
                        lastElapsedMilliseconds[index] = elapsedMilliseconds;
                    }
                };

                TestSection.Act();

                stopwatch.Start();
                // ReSharper disable once AssignNullToNotNullAttribute
                int handle = client.SubscribeDataChange(endpointDescriptor, nodeDescriptor, samplingInterval1);

                AssertUtilities.WaitingPhase(20 * samplingInterval1);

                AssertUtilities.WriteLine();
                float averageInterval1;
                lock (synchronize)
                {
                    Assert.AreNotEqual(0, intervalCount[0], "No interval 1");
                    averageInterval1 = (float)intervalSum[0] / intervalCount[0];
                }
                AssertUtilities.WriteLine($"Average interval 1: {averageInterval1}");
                Assert.IsTrue(samplingInterval1 / (1 + tolerance) <= averageInterval1, "Average interval 1 too short");
                Assert.IsTrue(averageInterval1 <= samplingInterval1 * (1 + tolerance), "Average interval 1 too long");

                client.ChangeMonitoredItemSubscription(handle, samplingInterval2);

                AssertUtilities.WaitingPhase(20 * samplingInterval2);

                AssertUtilities.WriteLine();
                float averageInterval2;
                lock (synchronize)
                {
                    Assert.AreNotEqual(0, intervalCount[1], "No interval 2");
                    averageInterval2 = (float)intervalSum[1] / intervalCount[1];
                }
                AssertUtilities.WriteLine($"Average interval 2: {averageInterval2}");
                Assert.IsTrue(samplingInterval2 / (1 + tolerance) <= averageInterval2, "Average interval 2 too short");
                Assert.IsTrue(averageInterval2 <= samplingInterval2 * (1 + tolerance), "Average interval 2 too long");

                client.UnsubscribeMonitoredItem(handle);

                TestSection.Assert();
            }
        }

        /// <summary>
        /// Delete an existing Subscription
        /// </summary>
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group006_SubscriptionServices_Test03_DeleteSubscription()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            const int samplingInterval = 1000;
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            UANodeDescriptor nodeDescriptor = DynamicNode1;
            AssertUtilities.WriteLine($"Node descriptor: {nodeDescriptor}");

            TestSection.Act();
            using (var client = new EasyUAClient
            {
                Isolated = true,
                IsolatedParameters = { SessionParameters = { EnableStatusSubscription = !SupportsJustOneSubscription } }
            })
            {
                var notificationList = new List<EasyUADataChangeNotificationEventArgs>();
                var synchronize = new object();
                client.DataChangeNotification += (sender, args) =>
                {
                    lock (synchronize)
                        notificationList.Add(args);
                };

                TestSection.Act();

                // ReSharper disable once AssignNullToNotNullAttribute
                int handle = client.SubscribeDataChange(endpointDescriptor, nodeDescriptor, samplingInterval);
                AssertUtilities.WaitingPhase(10 * samplingInterval);

                int capturedCount = notificationList.Count;
                AssertUtilities.WriteLine($"Captured count: {capturedCount}");
                Assert.IsTrue(capturedCount >= 2, "Not enough captured");

                client.UnsubscribeMonitoredItem(handle);

                AssertUtilities.WaitingPhase(3 * 1000);

                int oldCount = notificationList.Count;
                AssertUtilities.WriteLine($"Old count: {oldCount}");

                AssertUtilities.WaitingPhase(Math.Max(3 * 1000, 10 * oldCount));

                int newCount = notificationList.Count;
                AssertUtilities.WriteLine($"New count: {newCount}");

                TestSection.Assert();

                Assert.AreEqual(oldCount, newCount);

                foreach (EasyUADataChangeNotificationEventArgs notification in notificationList)
                {
                    Assert.IsNotNull(notification);
                    Assert.IsNull(notification.Exception);
                }

                // IMPROVE: test for actual success of the DeleteSubscription service
            }
        }

        /// <summary>
        /// SetPublishingMode() to enable/disable publishing. Test with multiple subscriptions if available.
        /// </summary>
        [TestMethod]
        public void Group006_SubscriptionServices_Test04_SetPublishingMode()
        {
            ClientNotSupported();
        }

        /// <summary>
        /// [obsolete] Publish Service to request attribute value changes
        /// </summary>
        [TestMethod]
        public void Group006_SubscriptionServices_Test05_Publish()
        {
            ClientNotSupported();
        }

        /// <summary>
        /// Publish() returns after keep-alive interval (no data-changes; hint: subscribe to static items).
        /// </summary>
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group006_SubscriptionServices_Test06_Publish_KeepAlive()
        {
            TestSection.Arrange();
            const int samplingInterval = 1000;
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient
            {
                Isolated = true,
                IsolatedParameters =
                {
                    SessionParameters = { EnableStatusSubscription = !SupportsJustOneSubscription } ,
                    SubscriptionParameters = { KeepAliveCount = 1 }
                }
            })
            {
                int failureNotificationCount = 0;
                int successNotificationCount = 0;
                var synchronize = new object();
                client.DataChangeNotification += (sender, args) =>
                {
                    lock (synchronize)
                        // ReSharper disable once PossibleNullReferenceException
                        if (args.Succeeded)
                            successNotificationCount++;
                        else
                            failureNotificationCount++;
                };

                TestSection.Act();

                // ReSharper disable once AssignNullToNotNullAttribute
                int handle = client.SubscribeDataChange(endpointDescriptor, StaticNode1, samplingInterval, 
                    UADataChangeTrigger.StatusValue);
                Thread.Sleep(10 * samplingInterval);
                client.UnsubscribeMonitoredItem(handle);

                int capturedFailureCount = failureNotificationCount;
                int capturedSuccessCount = successNotificationCount;
                AssertUtilities.WriteLine($"Captured failure count: {capturedFailureCount}");
                AssertUtilities.WriteLine($"Captured success count: {capturedSuccessCount}");

                TestSection.Assert();

                if ((capturedFailureCount == 0) && (capturedSuccessCount > 1))
                    Assert.Inconclusive("Possibly a dynamic node?");

                // If the keep-alive fails, there would be an additional event with the error
                Assert.AreEqual(1, capturedSuccessCount);
                Assert.AreEqual(0, capturedFailureCount);
            }
        }

        /// <summary>
        /// Issue multiple Publish() requests that are to be queued by the Server.
        /// </summary>
        [TestMethod]
        public void Group006_SubscriptionServices_Test07_Publish_Parallel()
        {
            ClientNotSupported();
        }

        /// <summary>
        /// [obsolete] Republish Service
        /// </summary>
        [TestMethod]
        public void Group006_SubscriptionServices_Test08_Republish()
        {
            ClientNotSupported();
        }

        /// <summary>
        /// Create subscriptions with varying number of monitored items and different rate settings as supported by the client. Verify that the server provides updates according to the specified publishing rate. Optionally use Write() service to trigger data changes.
        /// </summary>
        // IMPROVE: this test does not yet recognize that the server may revise the sampling rate
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void Group006_SubscriptionServices_Test09_DifferentRates()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UANodeDescriptor nodeDescriptor1 = DynamicNode1;
            Debug.Assert(nodeDescriptor1 != null, "Missing dynamic node 1");
            AssertUtilities.WriteLine($"Node descriptor 1: {nodeDescriptor1}");
            UANodeDescriptor nodeDescriptor2 = DynamicNode2;
            Debug.Assert(nodeDescriptor2 != null, "Missing dynamic node 2");
            AssertUtilities.WriteLine($"Node descriptor 2: {nodeDescriptor2}");
            int samplingInterval1 = DynamicSamplingInterval1;
            int samplingInterval2 = DynamicSamplingInterval2;
            const float tolerance = 0.6f;
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;
            AssertUtilities.WriteLine($"Sampling interval 1: {samplingInterval1}");
            AssertUtilities.WriteLine($"Sampling interval 2: {samplingInterval2}");
            AssertUtilities.WriteLine($"Tolerance: {tolerance}");

            TestSection.Act();
            using (var client = new EasyUAClient
            {
                Isolated = true,
                IsolatedParameters = { SessionParameters = { EnableStatusSubscription = !SupportsJustOneSubscription } }
            })
            {
                var stopwatch = new Stopwatch();
                long[] lastElapsedMilliseconds = { -1, -1 };
                long[] intervalSum = { 0, 0 };
                int[] intervalCount = { 0, 0 };
                var synchronize = new object();
                client.DataChangeNotification += (sender, args) =>
                {
                    long elapsedMilliseconds = stopwatch.ElapsedMilliseconds;
                    // ReSharper disable once PossibleNullReferenceException
                    int index = (int)args.Arguments.State;

                    AssertUtilities.WriteLine($"@{elapsedMilliseconds} data change {index + 1}: {args}");
                    lock (synchronize)
                    {
                        if (lastElapsedMilliseconds[index] != -1)
                        {
                            intervalSum[index] += elapsedMilliseconds - lastElapsedMilliseconds[index];
                            intervalCount[index]++;
                        }
                        lastElapsedMilliseconds[index] = elapsedMilliseconds;
                    }
                };

                TestSection.Act();

                stopwatch.Start();
                int[] handleArray = client.SubscribeMultipleMonitoredItems(new[]
                    {
                                new EasyUAMonitoredItemArguments(
                                    state: 0,
                                    endpointDescriptor:endpointDescriptor,
                                    nodeDescriptor:nodeDescriptor1,
                                    monitoringParameters:samplingInterval1),
                                new EasyUAMonitoredItemArguments(
                                    state: 1,
                                    endpointDescriptor:endpointDescriptor,
                                    nodeDescriptor:nodeDescriptor2,
                                    monitoringParameters:samplingInterval2),
                            });

                Thread.Sleep(20 * samplingInterval1);
                client.UnsubscribeMultipleMonitoredItems(handleArray);
                stopwatch.Stop();

                TestSection.Assert();

                float averageInterval1;
                float averageInterval2;
                lock (synchronize)
                {
                    Assert.AreNotEqual(0, intervalCount[0], "No interval 1");
                    Assert.AreNotEqual(0, intervalCount[1], "No interval 2");

                    averageInterval1 = (float)intervalSum[0] / intervalCount[0];
                    averageInterval2 = (float)intervalSum[1] / intervalCount[1];
                }

                AssertUtilities.WriteLine();
                AssertUtilities.WriteLine($"Average interval 1: {averageInterval1}");
                AssertUtilities.WriteLine($"Average interval 2: {averageInterval2}");

                Assert.IsTrue(samplingInterval1 / (1 + tolerance) <= averageInterval1, "Average interval 1 too short");
                Assert.IsTrue(averageInterval1 <= samplingInterval1 * (1 + tolerance), "Average interval 1 too long");

                Assert.IsTrue(samplingInterval2 / (1 + tolerance) <= averageInterval2, "Average interval 2 too short");
                Assert.IsTrue(averageInterval2 <= samplingInterval2 * (1 + tolerance), "Average interval 2 too long");

                // IMPROVE: "Optionally use Write() service to trigger data changes."
            }
        }

        #endregion
    }
}
