﻿
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Utilities;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.AddressSpace.Standard;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.Graphs;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.OpcInteroperabilityTest.UA
{
    partial class UAInteroperabilityWorkshopTest
    {
        #region Test Methods

        [TestMethod]
        public void Group002_ViewServices_Header____________________()
        {
        }

        /// <summary>
        /// Call Browse() on a single node in the address space. Repeat with other nodes.
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test01_Browse_SingleNode()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                UANodeElementCollection nodeElementCollection = client.BrowseDataNodes(endpointDescriptor);
                foreach (UANodeElement nodeElement in nodeElementCollection)
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"{nodeElement.DisplayName}: {nodeElement}");
                AssertUtilities.WriteLine($"Element count: {nodeElementCollection.Count}");

                TestSection.Assert();
                Assert.IsTrue(nodeElementCollection.Count >= 1, "Not enough elements");
            }
        }

        /// <summary>
        /// Call Browse() by specifying more than one node in a single call. Walk through the address space.
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test02_Browse_Walk()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;
            const int count = 100;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                AssertUtilities.Phase("Browse multiple nodes at once");

                UANodeElementCollectionResult[] nodeElementCollectionResults = client.BrowseMultiple(new []
                {
                    new UABrowseArguments(endpointDescriptor, UAObjectIds.Server, UABrowseParameters.Objects),
                    new UABrowseArguments(endpointDescriptor, UAObjectIds.TypesFolder, UABrowseParameters.Objects),
                });

                AssertUtilities.Phase("Address space walk");

                IEnumerable<UATreePosition> treePositions = client.TraverseNodeTree(endpointDescriptor);
                int i = 0;
                foreach (UATreePosition treePosition in treePositions.Take(count))
                {
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"[{i}] ({treePosition.Level}) {treePosition.DisplayName}: {treePosition}");
                    i++;
                }

                TestSection.Assert();

                AssertUtilities.Phase("Browse multiple nodes at once");
                Assert.IsNotNull(nodeElementCollectionResults);
                Assert.AreEqual(2, nodeElementCollectionResults.Length);
                UANodeElementCollectionResult nodeElementCollectionResult1 = nodeElementCollectionResults[0];
                Assert.IsNotNull(nodeElementCollectionResult1);
                Assert.IsTrue(nodeElementCollectionResult1.Succeeded);
                UANodeElementCollection nodeElementCollection1 = nodeElementCollectionResult1.Value;
                Assert.IsNotNull(nodeElementCollection1);
                Assert.IsTrue(nodeElementCollection1.Count >= 4);
                UANodeElementCollectionResult nodeElementCollectionResult2 = nodeElementCollectionResults[0];
                Assert.IsNotNull(nodeElementCollectionResult2);
                Assert.IsTrue(nodeElementCollectionResult2.Succeeded);
                UANodeElementCollection nodeElementCollection2 = nodeElementCollectionResult2.Value;
                Assert.IsNotNull(nodeElementCollection2);
                Assert.IsTrue(nodeElementCollection2.Count >= 4);

                AssertUtilities.Phase("Address space walk");
                Assert.IsTrue(i >= 1, "Not enough elements");
            }
        }

        /// <summary>
        /// Call Browse() using various Reference filters; e.g. only hierarchical references.
        /// </summary>
        // IMPROVE: do "various"
        [TestMethod]
        public void Group002_ViewServices_Test03_Browse_References()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                var browseParameters = new UABrowseParameters(UANodeClass.All, 
                    EnumerableUtilities.One(UAReferenceTypeIds.HierarchicalReferences));
                UANodeElementCollection nodeElementCollection = client.Browse(endpointDescriptor, 
                    UAObjectIds.ObjectsFolder, browseParameters);
                foreach (UANodeElement nodeElement in nodeElementCollection)
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"{nodeElement.DisplayName}: {nodeElement}");
                AssertUtilities.WriteLine($"Element count: {nodeElementCollection.Count}");

                TestSection.Assert();
                Assert.IsTrue(nodeElementCollection.Count >= 1, "Not enough elements");
            }
        }

        /// <summary>
        /// Call Browse() using various NodeClass filters (e.g. to return Variable nodes only).
        /// </summary>
        // IMPROVE: do "various"
        [TestMethod]
        public void Group002_ViewServices_Test04_Browse_NodeClasses()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                AssertUtilities.Phase("Browse 1");

                var browseParameters1 = new UABrowseParameters(UANodeClass.All);
                UANodeElementCollection nodeElementCollection1 = client.Browse(endpointDescriptor, 
                    UAObjectIds.ObjectsFolder, browseParameters1);
                foreach (UANodeElement nodeElement in nodeElementCollection1)
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"{nodeElement.DisplayName}: {nodeElement}; {nodeElement.NodeClass}");
                AssertUtilities.WriteLine($"Element count 1: {nodeElementCollection1.Count}");
                Assert.IsTrue(nodeElementCollection1.Count >= 1, "Not enough elements");

                // ReSharper disable once PossibleNullReferenceException
                UANodeClass nodeClassToExclude = nodeElementCollection1[0].NodeClass;
                AssertUtilities.WriteLine($"Node class to exclude: {nodeClassToExclude}");

                AssertUtilities.Phase("Browse 2");

                UANodeClass restrictedNodeClasses = UANodeClass.All & ~nodeClassToExclude;
                AssertUtilities.WriteLine($"Restricted node classes: {restrictedNodeClasses}");
                var browseParameters2 = new UABrowseParameters(restrictedNodeClasses);
                UANodeElementCollection nodeElementCollection2 = client.Browse(endpointDescriptor, UAObjectIds.ObjectsFolder, browseParameters2);
                foreach (UANodeElement nodeElement in nodeElementCollection2)
                    // ReSharper disable once PossibleNullReferenceException
                    AssertUtilities.WriteLine($"{nodeElement.DisplayName}: {nodeElement}; {nodeElement.NodeClass}");
                AssertUtilities.WriteLine($"Element count 2: {nodeElementCollection2.Count}");

                TestSection.Assert();

                Assert.IsTrue(nodeElementCollection2.Count < nodeElementCollection1.Count,
                    "Effectively restricting the node classes has not reduced the count of returned elements");
                // ReSharper disable once PossibleNullReferenceException
                Assert.IsTrue(nodeElementCollection2.All(element => (element.NodeClass & nodeClassToExclude) == 0),
                    "An element has been returned that has the excluded node class");
            }
        }

        /// <summary>
        /// Call Browse() with a filter to reduce the result information (e.g. omit BrowseName and ReferenceType).
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test05_Browse_ReduceResult()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        /// <summary>
        /// [obsolete] BrowseNext (Continuation Point)
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test06_BrowseNext_ContinuationPoint()
        {
            ClientNotSupported();
        }

        /// <summary>
        /// TranslateBrowsePathToNodeId to determine the NodeID of components (e.g. properties)
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test07_TranslateBrowsePathToNodeId_Components()
        {
            TestSection.Arrange();
            PrecheckTestMethod();
            UAEndpointDescriptor endpointDescriptor = EndpointDescriptorCommon;

            TestSection.Act();
            using (var client = new EasyUAClient { Isolated = true })
            {
                const int count = 100;
                IEnumerable<UATreePosition> treePositions = client.TraverseNodeTree(endpointDescriptor);
                UATreePosition variablePosition = treePositions
                    .Take(count)
                    // ReSharper disable once PossibleNullReferenceException
                    .FirstOrDefault(position => position.NodeClass == UANodeClass.Variable);
                Assert.IsNotNull(variablePosition, "Found no variables");

                AssertUtilities.WriteLine($"Browse path: {variablePosition.BrowsePath}");
                UANodeDescriptor nodeDescriptor = variablePosition.BrowsePath;

                object value = client.ReadValue(endpointDescriptor, nodeDescriptor);
                AssertUtilities.WriteLine($"Value: {value}");
            }
        }

        /// <summary>
        /// [obsolete] TranslateBrowsePathToNodeId for non-hierarchical References
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test08_TranslateBrowsePathToNodeId_NonHierarchical()
        {
            ClientNotSupported();
        }

        /// <summary>
        /// Register()/UnregisterNodes() for one or more standard UA-defined Nodes (e.g. "Server") in a single call.
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test09_RegisterUnregister_StandardNodes()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        /// <summary>
        /// Register()/UnregisterNodes() for one or more vendor-defined Nodes in a single call.
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test10_RegisterUnregister_CustomNodes()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        /// <summary>
        /// Browse with Views.
        /// </summary>
        [TestMethod]
        public void Group002_ViewServices_Test11_Browse_Views()
        {
            ClientNotSupported();   // IMPROVE: have this
        }

        #endregion
    }
}
